# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import logging
from collections import OrderedDict
from typing import Optional

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.asymmetric import ec
from pyasn1.codec.der import decoder as der_decoder
from pyasn1.codec.der import encoder as der_encoder
from pyasn1.codec.native.encoder import encode as native_encoder

from manifesttool.mtool import ecdsa_helper
from manifesttool.mtool.asn1.manifest_codec import ManifestAsnCodecBase
from manifesttool.mtool.asn1.v3 import manifest_schema_v3 as manifest_schema


class ManifestAsnCodecV3(ManifestAsnCodecBase):
    VERSION = 'v3'

    logger = logging.getLogger('v3-manifest-codec')

    def __init__(self, dom=None):

        if dom:
            self.dom = dom
        else:
            self.dom = manifest_schema.Manifest()

    @classmethod
    def get_name(cls) -> str:
        return cls.VERSION

    def set_payload_version(self, version: str):
        assert isinstance(version, str)
        self.dom['payload-version'] = version

    def set_update_priority(self, priority: int):
        self.dom['update-priority'] = priority

    def set_vendor_id(self, vendor_id: bytes):
        self.dom['vendor-id'] = vendor_id

    def set_vendor_data(self, data: bytes):
        self.dom['vendor-data'] = data

    def set_class_id(self, class_id: bytes):
        self.dom['class-id'] = class_id

    def set_device_id(self, device_id: bytes):
        self.dom['device-id'] = device_id

    def set_payload_fingerprint(self, digest: bytes, size: int):
        self.dom['payload-digest'] = digest
        self.dom['payload-size'] = size

    def set_payload_uri(self, uri: str):
        self.dom['payload-uri'] = uri

    def set_payload_format(self, payload_format: str):
        self.dom['payload-format'] = payload_format

    def set_payload_metadata(
            self,
            installed_digest: bytes,
            installed_size: int,
            precursor_digest: bytes
    ):
        delta_meta_dom = manifest_schema.PayloadMetadata()
        delta_meta_dom['installed-size'] = installed_size
        delta_meta_dom['installed-digest'] = installed_digest
        if precursor_digest:
            delta_meta_dom['precursor-digest'] = precursor_digest
        self.dom['payload-metadata'] = delta_meta_dom

    @classmethod
    def decode(
            cls,
            data: bytes,
            verification_key: Optional[ec.EllipticCurvePublicKey]
    ) -> OrderedDict:

        signed_resource_dom = der_decoder.decode(
            data, asn1Spec=manifest_schema.SignedResource())[0]
        if not verification_key:
            return native_encoder(signed_resource_dom)

        codec = ManifestAsnCodecV3(dom=signed_resource_dom['manifest'])
        signed_data = codec.get_signed_data()
        der_signature = bytes(signed_resource_dom['signature'])
        signature = ecdsa_helper.signature_raw_to_der(der_signature)

        try:
            ecdsa_helper.ecdsa_verify(
                public_key=verification_key,
                signed_data=signed_data,
                signature=signature
            )
            cls.logger.info('Manifest Signature verified!')
        except InvalidSignature as ex:
            raise AssertionError(
                'Manifest Signature verification failed') from ex

        raw_image_signature = bytes(
            codec.dom['installed-signature'])
        if raw_image_signature:
            der_image_signature = \
                ecdsa_helper.signature_raw_to_der(
                    raw_image_signature
                )
            digest = bytes(codec.dom['payload-digest'])
            if str(codec.dom['payload-format']) == \
                    'arm-patch-stream':
                digest = bytes(
                    codec.dom['payload-metadata']['installed-digest']
                )
            try:
                ecdsa_helper.ecdsa_verify_prehashed(
                    public_key=verification_key,
                    digest=digest,
                    signature=der_image_signature
                )
                cls.logger.info('Image Signature verified!')
            except InvalidSignature as ex:
                raise AssertionError(
                    'Image Signature verification failed') from ex

        return native_encoder(signed_resource_dom)

    def get_signed_data(self) -> bytes:
        # sha_content = utils.sha_hash(enc_data)

        return der_encoder.encode(self.dom)

    def get_der_signed_resource(self, signature: bytes) -> bytes:
        resource_dom = manifest_schema.SignedResource()
        resource_dom['manifest-version'] = self.VERSION
        resource_dom['manifest'] = self.dom
        resource_dom['signature'] = signature
        return der_encoder.encode(resource_dom)

    def set_update_certificate(self, cert_data: bytes):
        raise NotImplementedError

    def set_image_signature(self, signature: bytes):
        self.dom['installed-signature'] = signature

    def set_component_name(self, component: str):
        self.dom['component-name'] = component



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import argparse
import binascii
import collections
import logging

import yaml
from pyasn1.error import PyAsn1Error

from manifesttool.mtool import ecdsa_helper
from manifesttool.mtool.asn1 import ManifestVersion


def bytes_representer(dumper: yaml.Dumper, obj):
    try:
        return dumper.represent_str(obj.decode('utf-8'))
    except UnicodeError:
        return dumper.represent_str(binascii.hexlify(obj).decode('utf-8'))


def ordered_dict_representer(dumper, data):
    return dumper.represent_dict(data.items())


yaml.Dumper.add_representer(collections.OrderedDict, ordered_dict_representer)
yaml.Dumper.add_representer(bytes, bytes_representer)

class ParseAction:

    logger = logging.getLogger('manifest-tool-parse')

    @staticmethod
    def register_parser_args(parser):
        required = parser.add_argument_group('required arguments')
        optional = parser.add_argument_group('optional arguments')

        required.add_argument(
            'manifest',
            help='Path to the manifest file.',
            type=argparse.FileType('rb')
        )

        key_or_cert_group = optional.add_mutually_exclusive_group()
        key_or_cert_group.add_argument(
            '-c', '--certificate',
            help='Path to a certificate file to validate '
                 'the manifest signature.',
            type=argparse.FileType('rb')
        )
        key_or_cert_group.add_argument(
            '-p', '--public-key',
            help='Path to a public key file, '
                 'containing a key in uncompressed point format, '
                 'to validate the manifest signature.',
            type=argparse.FileType('rb')
        )
        key_or_cert_group.add_argument(
            '-k', '--private-key',
            help='Path to a private key PEM file '
                 'to validate the manifest signature.',
            type=argparse.FileType('rb')
        )

        optional.add_argument(
            '-h',
            '--help',
            action='help',
            help='Show this help message and exit.'
        )

    # pylint: disable=too-many-branches
    @classmethod
    def do_parse(
            cls,
            manifest_data: bytes,
            certificate_data: bytes,
            public_key_data: bytes,
            private_key_data: bytes
    ):
        dom = None

        public_key = None
        if private_key_data:
            public_key = ecdsa_helper.public_key_from_private(private_key_data)
        elif certificate_data:
            try:
                public_key = ecdsa_helper.public_key_from_certificate(
                    certificate_data)
            except ValueError as ex:
                raise AssertionError('Malformed certificate') from ex
        elif public_key_data:
            public_key = ecdsa_helper.public_key_from_bytes(public_key_data)

        for codec_class in ManifestVersion.list_codecs():
            try:
                dom = codec_class.decode(
                    manifest_data, public_key)
                break
            except PyAsn1Error:
                continue
        if not dom:
            raise AssertionError('Malformed manifest')

        logging.info(
            '\n----- Manifest dump start -----\n'
            '%s----- Manifest dump end -----',
            yaml.dump(dom, default_flow_style=False)

        )

    @classmethod
    def entry_point(cls, args):
        cert_data = args.certificate.read() if args.certificate else None
        public_key = args.public_key.read() if args.public_key else None
        private_key = args.private_key.read() if args.private_key else None
        manifest_data = args.manifest.read()

        cls.do_parse(
            manifest_data=manifest_data,
            certificate_data=cert_data,
            public_key_data=public_key,
            private_key_data=private_key
        )



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import argparse
import re
from pathlib import Path
from typing import List, Tuple

def non_negative_int_arg_factory(value: str) -> int:
    """
    Construct non negative integer value for an argument
    :param value: input string
    :return: integer value
    """
    prospective = None
    try:
        prospective = int(value)
    except ValueError:
        pass
    if prospective is None or prospective < 0:
        raise argparse.ArgumentTypeError(
            '"{}" is an invalid non-negative integer value'.format(value))
    return prospective

def semantic_version_arg_factory(value) -> str:
    """
    Construct major, minor, split tuple for an argument
    :param value: input string
    :return: str
    """
    nibble = '([0-9]|[1-9][0-9]{0,2})'
    pattern = '^' + r'\.'.join([nibble, nibble, nibble]) + '$'
    if not re.match(pattern, value) or value == '0.0.0':
        raise argparse.ArgumentTypeError(
            '"{}" is not a valid SemVer format. '
            'Min. 0.0.1, max 999.999.999'.format(value))
    return value

def _semver_to_int(value: str) -> int:
    """
    Helper to construct integer value for SemVer string
    :param value: input string in SemVer format
    :return: integer value
    """
    # 'x.y.z' -> 0x008xxx000yyy0zzz
    # Note, result must be same as client
    #  fota_component_version_semver_to_int
    major_bits = 16 + 24
    minor_bits = 16
    split_bits = 0
    major, minor, split = value.split('.')
    return (1 << 55) | \
           (int(major) << major_bits) | \
           (int(minor) << minor_bits) | \
           (int(split) << split_bits)

def semver_as_tuple_arg_factory(value: str) -> Tuple[int, str]:
    """
    Construct major, minor, split tuple for an argument
    :param value: input string
    :return: tuple of integer value and original str value
    """
    value = semantic_version_arg_factory(value)
    return _semver_to_int(value), value

def existing_file_path_arg_factory(value):
    """
        Construct Path to an existing file for an argument
        :param value: input string
        :return: major, minor, split tuple
    """
    prospective = Path(value)
    if not prospective.is_file():
        raise argparse.ArgumentTypeError(
            'File "{}" is not found'.format(value)
        )
    return prospective



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import argparse
import datetime
import logging
import time
import uuid
from pathlib import Path
import shutil

import yaml
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec

from manifesttool import __version__
from manifesttool.dev_tool import defaults
from manifesttool.mtool import ecdsa_helper

SCRIPT_DIR = Path(__file__).resolve().parent

logger = logging.getLogger('manifest-dev-tool-init')


def uuid_factory(value: str) -> uuid.UUID:
    """
    Construct UUID from a string command line argument argument
    :param value: input string
    :return: UUID
    """
    try:
        return uuid.UUID(hex=value)
    except ValueError as exc:
        raise argparse.ArgumentTypeError(
            '"{}" is a malformed hexadecimal UUID string'.format(value)
        ) from exc


class DeprecateStoreTrueAction(argparse.Action):  # pylint: disable=R0903
    def __init__(self,
                 option_strings,
                 dest,
                 required=False,
                 help=None):  # pylint: disable=W0622
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=0,
            const=True,
            default=False,
            required=required,
            help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        logger.warning(
            '%s option is deprecated and has no effect', option_string)
        setattr(namespace, self.dest, self.const)

def register_parser(parser: argparse.ArgumentParser):
    parser.add_argument(
        '-h',
        '--help',
        action='help',
        help='Show this help message and exit.'
    )
    parser.add_argument(
        '-f', '--force',
        action=DeprecateStoreTrueAction,
        help='Deprecated. Has no effect. Will be removed in future versions'
    )

    parser.add_argument(
        '--cache-dir',
        help='Tool\'s cache directory for preserving the state between '
             '"init", "create" and "update" command invocations. '
             '[Default: {}]'.format(defaults.BASE_PATH),
        type=Path,
        default=defaults.BASE_PATH
    )

    service = parser.add_argument_group(
        'optional arguments (Pelion Device Management service configuration)')
    service.add_argument(
        '-p', '--gw-preset',
        help='The preset name defined in {}, '
             ' which specifies a URL and '
             'access key.'.format(defaults.PELION_GW_PATH),
        choices=defaults.PELION_GW.keys() if defaults.PELION_GW else []
    )
    service.add_argument(
        '-a', '--access-key', '--api-key',
        help='Access key for accessing a Pelion Device Management API.'
    )
    service.add_argument(
        '-u', '--api-url',
        help='Pelion Device Management API URL. '
             '[Default: {}]'.format(defaults.API_GW),
        default=defaults.API_GW
    )

    parser.add_argument(
        '-g', '--generated-resource',
        help='Generated update resource C filename. '
             '[Default: {}]'.format(defaults.UPDATE_RESOURCE_C),
        type=Path,
        default=defaults.UPDATE_RESOURCE_C
    )

    parser.add_argument(
        '--vendor-id',
        help='Set custom vendor UUID. [Default: random UUID]',
        type=uuid_factory,
        default=uuid.uuid4()
    )
    parser.add_argument(
        '--class-id',
        help='Set custom class UUID. [Default: random UUID]',
        type=uuid_factory,
        default=uuid.uuid4()
    )

    parser.add_argument(
        '--update-certificate',
        help='Path to the update certificate file.',
        type=Path
    )
    parser.add_argument(
        '--key',
        help='Path to the PEM format private key file.',
        type=Path
    )


def chunks(arr, num_of_elements):
    for i in range(0, len(arr), num_of_elements):
        yield arr[i:i + num_of_elements]


def format_rows(arr):
    for _slice in chunks(arr, 8):
        yield ', '.join(['0x{:02X}'.format(x) for x in _slice])


def pretty_print(arr):
    return ',\n    '.join(list(format_rows(arr)))


def generate_update_default_resources_c(
        c_source: Path,
        vendor_id: uuid.UUID,
        class_id: uuid.UUID,
        private_key_file: Path,
        cert_file: Path
):
    """
    Generate update resources C source file for developer convenience.

    :param c_source: generated C source file
    :param vendor_id: vendor UUID
    :param class_id: class UUID
    :param private_key_file: private key file
    :param cert_file: update certificate file
    """

    vendor_id_str = pretty_print(vendor_id.bytes)
    class_id_str = pretty_print(class_id.bytes)

    cert_data = cert_file.read_bytes()

    cert_str = pretty_print(cert_data)

    template = (SCRIPT_DIR / 'code_template.txt').read_text()

    public_key = ecdsa_helper.public_key_from_private(
        private_key_file.read_bytes())
    public_key_bytes = ecdsa_helper.public_key_to_bytes(public_key)
    update_public_key_str = pretty_print(public_key_bytes)

    c_source.write_text(
        template.format(
            vendor_id=vendor_id_str,
            class_id=class_id_str,
            cert=cert_str,
            timestamp=time.strftime("%Y-%m-%d %H:%M:%S %Z"),
            tool='manifest-dev-tool init',
            version=__version__,
            update_pub_key=update_public_key_str
        )
    )
    logger.info('generated update source %s', c_source)


def import_credentials(
        origin_key_file: Path,
        origin_cert_file: Path,
        dest_key_file: Path,
        dest_cert_file: Path
):
    """
    Import developer credentials

    :param origin_key_file - path to .pem signing key file
    :param origin_cert_file - path to .der public key certificate file
    :param dest_key_file - destination file path
    :param dest_cert_file - destination file path
    """

    logger.info('importing dev-credentials')

    if origin_key_file != dest_key_file:
        shutil.copy(origin_key_file, dest_key_file)
        logger.info(
            'imported %s to %s',
            origin_key_file,
            dest_key_file
        )

    if origin_cert_file != dest_cert_file:
        shutil.copy(origin_cert_file, dest_cert_file)
        logger.info(
            'imported %s to %s',
            origin_cert_file,
            dest_cert_file
        )

    logger.info('dev-credentials - imported')


def generate_credentials(
        key_file: Path,
        cert_file: Path,
        cred_valid_time: int
):
    """
    Generate developer credentials

    :param key_file - path to .pem signing key file
    :param cert_file - path to .der public key certificate file
    :param cred_valid_time - x.509 certificate validity period
    """

    try:
        logger.info('generating dev-credentials')
        key = ec.generate_private_key(ec.SECP256R1(), default_backend())

        # For a self-signed certificate the
        # subject and issuer are always the same.
        subject_list = [
            x509.NameAttribute(x509.oid.NameOID.COMMON_NAME, 'localhost')]

        subject = issuer = x509.Name(subject_list)

        subject_key = key.public_key().public_bytes(
            serialization.Encoding.DER,
            serialization.PublicFormat.SubjectPublicKeyInfo
        )
        hash_ctx = hashes.Hash(hashes.SHA256(), backend=default_backend())
        hash_ctx.update(subject_key)
        key_digest = hash_ctx.finalize()

        subject_key_identifier = key_digest[:160 // 8]  # Use RFC7093, Method 1

        cert = x509.CertificateBuilder().subject_name(
            subject
        ).issuer_name(
            issuer
        ).public_key(
            key.public_key()
        ).serial_number(
            x509.random_serial_number()
        ).not_valid_before(
            datetime.datetime.utcnow()
        ).not_valid_after(
            datetime.datetime.utcnow() + datetime.timedelta(
                days=cred_valid_time)
        ).add_extension(
            x509.KeyUsage(
                digital_signature=True,
                content_commitment=False,
                key_encipherment=False,
                data_encipherment=False,
                key_agreement=False,
                key_cert_sign=False,
                crl_sign=False,
                encipher_only=False,
                decipher_only=False),
            critical=False
        ).add_extension(
            x509.SubjectAlternativeName([x509.DNSName(u"localhost")]),
            critical=False,
        ).add_extension(
            x509.ExtendedKeyUsage([x509.oid.ExtendedKeyUsageOID.CODE_SIGNING]),
            critical=False,
        ).add_extension(
            x509.SubjectKeyIdentifier(subject_key_identifier),
            critical=False,
            # Sign our certificate with our private key
        ).sign(key, hashes.SHA256(), default_backend())

        key_file.parent.mkdir(parents=True, exist_ok=True)
        key_file.write_bytes(
            key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption()
            )
        )
        logger.info('created %s', key_file)

        cert_file.parent.mkdir(parents=True, exist_ok=True)
        cert_file.write_bytes(
            cert.public_bytes(serialization.Encoding.DER))

        logger.info('created %s', cert_file)

        logger.info('dev-credentials - generated')

    except ValueError:
        logger.error('failed to generate dev-credentials')

        if key_file.is_file():
            key_file.unlink()
        if cert_file.is_file():
            cert_file.unlink()
        raise


def generate_developer_config(
        key_file: Path,
        cert_file: Path,
        config: Path,
        vendor_id: uuid.UUID,
        class_id: uuid.UUID
):

    cfg_data = {
        'key_file': key_file.as_posix(),
        'certificate': cert_file.as_posix(),
        'class-id': class_id.bytes.hex(),
        'vendor-id': vendor_id.bytes.hex(),
    }

    config.parent.mkdir(parents=True, exist_ok=True)
    with config.open('wt') as fh:
        yaml.dump(cfg_data, fh)

    logger.info('generated developer config file %s', config)


def generate_service_config(
    access_key: str,
    api_url: str,
    api_config_path: Path
):
    cfg = dict()

    if access_key is None or api_url is None:
        return

    if api_config_path.is_file():
        with api_config_path.open('rt') as fh:
            cfg = yaml.safe_load(fh)
            # if generated with previous versions
            # delete api_key
            if 'api_key' in cfg:
                cfg.pop('api_key')

    cfg['access_key'] = access_key
    cfg['host'] = api_url

    with api_config_path.open('wt') as fh:
        yaml.safe_dump(cfg, fh)


def entry_point(
        args,
        parser: argparse.ArgumentParser
):

    cache_dir = args.cache_dir
    cache_dir.mkdir(parents=True, exist_ok=True)

    vendor_id = args.vendor_id
    class_id = args.class_id

    import_key = getattr(args, 'key', None)
    import_cert = getattr(args, 'update_certificate', None)

    if import_key or import_cert:
        if not import_key or not import_cert:
            parser.error(
                'require --key and --update-certificate '
                'or none of those.'
            )
        import_credentials(
            origin_key_file=import_key,
            origin_cert_file=import_cert,
            dest_key_file=cache_dir / defaults.UPDATE_PRIVATE_KEY,
            dest_cert_file=cache_dir / defaults.UPDATE_PUBLIC_KEY_CERT
        )
    else:
        generate_credentials(
            key_file=cache_dir / defaults.UPDATE_PRIVATE_KEY,
            cert_file=cache_dir / defaults.UPDATE_PUBLIC_KEY_CERT,
            cred_valid_time=365 * 20  # years
        )

    generate_developer_config(
        key_file=cache_dir / defaults.UPDATE_PRIVATE_KEY,
        cert_file=cache_dir / defaults.UPDATE_PUBLIC_KEY_CERT,
        config=cache_dir / defaults.DEV_CFG,
        vendor_id=vendor_id,
        class_id=class_id
    )

    generate_update_default_resources_c(
        c_source=args.generated_resource,
        vendor_id=vendor_id,
        class_id=class_id,
        private_key_file=cache_dir / defaults.UPDATE_PRIVATE_KEY,
        cert_file=cache_dir / defaults.UPDATE_PUBLIC_KEY_CERT
    )
    access_key = args.access_key
    if not access_key and hasattr(args, 'gw_preset') and args.gw_preset:
        access_key = \
            defaults.PELION_GW[args.gw_preset].get('access_key')
        if not access_key:
            access_key = \
                defaults.PELION_GW[args.gw_preset].get('api_key')

    api_url = args.api_url
    if hasattr(args, 'gw_preset') and args.gw_preset:
        api_url = defaults.PELION_GW[args.gw_preset].get('host')

    generate_service_config(
        access_key=access_key,
        api_url=api_url,
        api_config_path=cache_dir / defaults.CLOUD_CFG
    )

    if not access_key:
        logger.warning('Access key not provided, so assisted '
                       'campaign management will not be available')

    (cache_dir / defaults.DEV_README).write_text(
        'Files in this directory are autogenerated by "manifest-dev-tool init"'
        ' tool for internal use only.\nDo not modify.\n'
    )



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import argparse
import enum
import logging
import sys

from manifesttool import __version__
from manifesttool.dev_tool.actions import create
from manifesttool.dev_tool.actions import init, update
from manifesttool.mtool.asn1.v1 import ManifestAsnCodecV1
from manifesttool.mtool.asn1.v3 import ManifestAsnCodecV3

logger = logging.getLogger('manifest-dev-tool')


class DevActions(enum.Enum):
    INIT = 'init'

    CREATE = 'create'
    CREATE_V1 = 'create-v1'

    UPDATE = 'update'
    UPDATE_V1 = 'update-v1'


def get_parser():
    parser = argparse.ArgumentParser(
        description='Developer tool to simplify creating and '
                    'running update campaigns.',
        add_help=False
    )

    actions_parser = parser.add_subparsers(title='Commands', dest='action')
    actions_parser.required = True

    init_parser = actions_parser.add_parser(
        DevActions.INIT.value,
        help='Initialize the developer environment.',
        description='Initialize the developer environment.',
        add_help=False
    )
    init.register_parser(init_parser)

    create_parser = actions_parser.add_parser(
        DevActions.CREATE.value,
        help='Create a manifest.',
        description='Create a manifest.',
        add_help=False
    )
    create.register_parser(create_parser, ManifestAsnCodecV3.get_name())

    create_parser = actions_parser.add_parser(
        DevActions.CREATE_V1.value,
        help='Create a V1 schema manifest.',
        description='Create a V1 schema manifest.',
        add_help=False
    )
    create.register_parser(create_parser, ManifestAsnCodecV1.get_name())

    update_parser = actions_parser.add_parser(
        DevActions.UPDATE.value,
        help='Create and run update campaign.',
        description='Create and run update campaign.',
        add_help=False
    )
    update.register_parser(update_parser, ManifestAsnCodecV3.get_name())

    update_parser = actions_parser.add_parser(
        DevActions.UPDATE_V1.value,
        help='Create and run V1 schema update campaign.',
        description='Create and run V1 schema update campaign.',
        add_help=False
    )
    update.register_parser(update_parser, ManifestAsnCodecV1.get_name())

    parser.add_argument(
        '--debug',
        action='store_true',
        help='Use debug logging level and print exception info upon exiting.',
    )

    parser.add_argument(
        '-h',
        '--help',
        action='help',
        help='Show this help message and exit.'
    )
    parser.add_argument(
        '--version',
        action='version',
        version='Manifest-Tool version {}'.format(__version__),
        help='Show program\'s version number and exit.'
    )

    return parser


def entry_point(argv=sys.argv[1:]):  # pylint: disable=dangerous-default-value
    logging.basicConfig(
        stream=sys.stdout,
        format='%(asctime)s %(levelname)s %(message)s',
        level=logging.INFO
    )
    parser = get_parser()
    args = parser.parse_args(argv)

    if args.debug:
        root = logging.getLogger()
        root.setLevel(logging.DEBUG)

    try:
        action = DevActions(args.action)

        if action == DevActions.INIT:
            init.entry_point(args, parser)
        # ---------------------------------------------------------------------
        elif action == DevActions.CREATE:
            create.entry_point(args, parser, ManifestAsnCodecV3)
        elif action == DevActions.CREATE_V1:
            create.entry_point(args, parser, ManifestAsnCodecV1)
        # ---------------------------------------------------------------------
        elif action == DevActions.UPDATE:
            update.entry_point(args, ManifestAsnCodecV3)
        elif action == DevActions.UPDATE_V1:
            update.entry_point(args, ManifestAsnCodecV1)
        # ---------------------------------------------------------------------
        else:
            raise AssertionError('unknown action')
    except Exception as ex:  # pylint: disable=broad-except
        logger.error(
            str(ex),
            exc_info=args.debug
        )
        return 1
    return 0


if __name__ == '__main__':
    raise SystemExit(entry_point())



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import argparse
import logging
import time
from pathlib import Path
from typing import Type

import yaml
import requests

from manifesttool.dev_tool import defaults
from manifesttool.dev_tool.actions.create \
    import load_cfg_and_get_fw_ver
from manifesttool.dev_tool.actions.create \
    import register_parser as register_create_parser
from manifesttool.dev_tool.actions.create import create_dev_manifest
from manifesttool.dev_tool.pelion import pelion
from manifesttool.mtool.actions import non_negative_int_arg_factory
from manifesttool.mtool.asn1 import ManifestAsnCodecBase

logger = logging.getLogger('manifest-dev-tool-update')

# The update API has a maximum name length of 128, but this is not queryable.
MAX_NAME_LEN = 128


def register_parser(parser: argparse.ArgumentParser, schema_version: str):

    register_create_parser(parser, schema_version, True)

    pdm_group = parser.add_argument_group(
        'optional arguments (Tool\'s campaign management)'
    )

    pdm_group.add_argument(
        '-i', '--device-id',
        help='Device ID for targeting a specific device in '
             'update campaign filter.'
    )

    pdm_group.add_argument(
        '-s', '--start-campaign',
        action='store_true',
        help='Create and start update campaign automatically.'
    )
    pdm_group.add_argument(
        '-w', '--wait-for-completion',
        dest='wait',
        action='store_true',
        help='Create and start update campaign automatically, '
             'wait for it to finish, and clean up created resources.'
    )
    pdm_group.add_argument(
        '-t', '--timeout',
        type=non_negative_int_arg_factory,
        help='Wait timeout in seconds. '
             '[Default: 0 - wait-forever]',
        default=0
    )
    pdm_group.add_argument(
        '-n', '--no-cleanup',
        action='store_true',
        help='Skip cleaning up created resources. '
             'Only relevant if --wait-for-completion was provided.'
    )

def _manage_campaign(api: pelion.UpdateServiceApi,
                     campaign_id: pelion.ID,
                     end_time: int,
                     do_wait: bool):
    starting = True
    old_state = ''
    try:
        api.campaign_start(campaign_id)
    except requests.HTTPError as ex:
        logger.error('Failed to start campaign %s', campaign_id)
        raise
    try:
        while end_time == 0 or time.time() < end_time:
            try:
                curr_campaign = api.campaign_get(campaign_id)
            except requests.ConnectionError as ex:
                logger.debug('Connection error %s', ex)
                time.sleep(3)
                continue
            curr_state = curr_campaign['state']
            curr_phase = curr_campaign['phase']
            if starting:
                if api.campaign_is_not_started(curr_phase):
                    raise AssertionError(
                        'Campaign not started - check filter and campaign '
                        'state.\n'
                        'Reason: {}'.format(curr_campaign['autostop_reason']))
                if api.campaign_is_active(curr_phase):
                    logger.info('Started Campaign ID: %s', campaign_id)
                    starting = False
                    if not do_wait:
                        return
            else:
                if not api.campaign_is_active(curr_phase):
                    logger.info(
                        "Campaign is finished in state: %s", curr_state)
                    return
            if old_state != curr_state:
                logger.info("Campaign state: %s", curr_state)
                old_state = curr_state
            time.sleep(3)
        logger.error('Campaign timed out')
        raise AssertionError('Campaign timed out')
    except requests.HTTPError as ex:
        logger.error('Failed to retrieve campaign state')
        raise

def _print_summary(summary: dict, summary_reasons: dict):
    logger.info("----------------------------")
    logger.info("    Campaign Summary ")
    logger.info("----------------------------")
    logger.info(" Successfully updated:   %d",
                summary.get('success') if summary.get('success') else 0)
    logger.info(" Failed to update:       %d",
                summary.get('fail') if summary.get('fail') else 0)
    logger.info(" Skipped:                %d",
                summary.get('skipped') if summary.get('skipped') else 0)
    logger.info(" Pending:                %d",
                summary.get('info') if summary.get('info') else 0)
    logger.info(" Total in this campaign: %d", sum(summary.values()))
    if summary_reasons.get('fail'):
        logger.warning("Reasons for failed updates:")
        for reasons in summary_reasons['fail']:
            if reasons.get('description'):
                logger.warning(" %s", reasons.get('description'))
    if summary_reasons.get('skipped'):
        logger.warning("Reasons for skipped updates:")
        for reasons in summary_reasons['skipped']:
            if reasons.get('description'):
                logger.warning(" %s", reasons.get('description'))


def _finalize(api: pelion.UpdateServiceApi,
              do_cleanup: bool,
              campaign_id: pelion.ID,
              manifest_id: pelion.ID,
              fw_image_id: pelion.ID,
              manifest_path: Path):
    summary = {}
    summary_reasons = {}
    failed_devices = []
    try:
        if campaign_id:
            # stop the campaign if it's still active
            api.campaign_stop(campaign_id)
            # get summary, failed devices and reasons
            statistics = api.campaign_statistics(campaign_id)
            summary = {s['id']: s['count'] for s in statistics}
            if summary.get('fail') and summary.get('fail') > 0:
                devices_state = api.campaign_device_metadata(campaign_id)
                failed_devices = [d['device_id'] for d in devices_state
                                  if d['deployment_state'] != 'deployed']
                summary_reasons['fail'] = \
                    api.campaign_statistic_events(campaign_id, 'fail')
            if summary.get('skipped') and summary.get('skipped') > 0:
                summary_reasons['skipped'] = \
                    api.campaign_statistic_events(campaign_id, 'skipped')
        if do_cleanup:
            logger.info('Cleaning up resources...')
            if campaign_id:
                api.campaign_delete(campaign_id)
            if manifest_id:
                api.manifest_delete(manifest_id)
            if fw_image_id:
                api.fw_delete(fw_image_id)
            if manifest_path and manifest_path.is_file():
                manifest_path.unlink()
    except requests.HTTPError as ex:
        logger.error('Failed to finalize update campaign')
        logger.debug('Exception %s', ex, exc_info=True)

    # print campaign summary
    if summary:
        _print_summary(summary, summary_reasons)
        if failed_devices:
            # assert if not all devices were updated
            raise AssertionError(
                'Failed to update {} devices: {}'.format(
                    len(failed_devices), ', '.join(failed_devices)))


def update(
        payload_path: Path,
        dev_cfg: dict,
        manifest_version: Type[ManifestAsnCodecBase],
        priority: int,
        vendor_data: Path,
        device_id: str,
        do_wait: bool,
        do_start: bool,
        end_time: int,
        do_cleanup: bool,
        service_config: Path,
        fw_version: str,
        sign_image: bool,
        component: str,
        short_url: bool,
        encrypt_payload: bool
):
    config = load_service_config(service_config)

    api = pelion.UpdateServiceApi(
        host=config['host'], access_key=config['access_key'])
    manifest_path = None
    fw_image_id = None
    manifest_id = None
    campaign_id = None
    encrypted_digest = None
    encrypted_size = None
    try:
        timestamp = time.strftime('%Y_%m_%d-%H_%M_%S')
        logger.info('Uploading FW image %s', payload_path.as_posix())

        fw_meta = api.fw_upload(
            fw_name='{}-{}'.format(timestamp, payload_path.name),
            image=payload_path,
            encrypt=encrypt_payload
        )

        fw_image_id = fw_meta['id']
        payload_url = fw_meta['short_datafile' if short_url else 'datafile']

        if encrypt_payload:
            if fw_meta.get('datafile_encryption', False) and \
               fw_meta.get('encrypted_datafile_checksum', False) and \
               fw_meta.get('encrypted_datafile_size', False):
                encrypted_digest = \
                    fw_meta['encrypted_datafile_checksum']
                encrypted_size = \
                    fw_meta['encrypted_datafile_size']
            else:
                raise AssertionError(
                    'Request to encrypt payload failed! '
                    'Device Management doesn\'t support this feature.'
                )

        manifest_data = create_dev_manifest(
            dev_cfg=dev_cfg,
            manifest_version=manifest_version,
            vendor_data_path=vendor_data,
            payload_path=payload_path,
            payload_url=payload_url,
            encrypted_digest=encrypted_digest,
            encrypted_size=encrypted_size,
            priority=priority,
            fw_version=fw_version,
            sign_image=sign_image,
            component=component
        )

        manifest_name = 'manifest-{timestamp}-{filename}'.format(
            filename=payload_path.name,
            timestamp=timestamp)
        manifest_path = payload_path.parent / manifest_name

        manifest_path.write_bytes(manifest_data)
        manifest_id = api.manifest_upload(
            name=manifest_name,
            manifest=manifest_path
        )

        campaign_name = 'campaign-{timestamp}-{filename}'.format(
            filename=payload_path.name,
            timestamp=timestamp)

        filters = [
            'device_class__eq={}'.format(dev_cfg['class-id']),
            'vendor_id__eq={}'.format(dev_cfg['vendor-id']),
        ]
        if device_id:
            filters.append('id__eq={}'.format(device_id))
        filters_query = '&'.join(filters)
        campaign_id = api.campaign_create(
            name=campaign_name,
            manifest_id=manifest_id,
            device_filter=filters_query
        )

        if do_start:
            _manage_campaign(api, campaign_id, end_time, do_wait)

    except requests.HTTPError as ex:
        # log additional service info
        srv_message = ex.response.json().get('message', None)
        if srv_message:
            logger.error(srv_message)
        raise
    except KeyboardInterrupt:
        logger.error('Aborted by user...')
    finally:
        if do_wait:
            _finalize(api,
                      do_cleanup,
                      campaign_id,
                      manifest_id,
                      fw_image_id,
                      manifest_path)


def load_service_config(service_config):
    if service_config.is_file():
        with service_config.open('rt') as fh:
            config = yaml.safe_load(fh)
            if 'host' in config and 'access_key' in config:
                return config
            if 'host' in config and 'api_key' in config:
                # replace api_key with access_key for backward compatibility
                config['access_key'] = config.pop('api_key')
                return config
    raise AssertionError('Pelion service configurations (URL and access key) '
                         'are not provided for assisted campaign '
                         'management')


def entry_point(
        args,
        manifest_version: Type[ManifestAsnCodecBase]
):
    dev_cfg, fw_version = \
        load_cfg_and_get_fw_ver(args, manifest_version)

    encrypt_payload = \
        hasattr(args, 'encrypt_payload') and args.encrypt_payload

    update(
        payload_path=args.payload_path,
        dev_cfg=dev_cfg,
        manifest_version=manifest_version,
        priority=args.priority,
        vendor_data=args.vendor_data,
        device_id=args.device_id,
        do_start=args.start_campaign or args.wait,
        do_wait=args.wait,
        end_time=time.time() + args.timeout if args.timeout > 0 else 0,
        do_cleanup=not args.no_cleanup,
        service_config=args.cache_dir / defaults.CLOUD_CFG,
        fw_version=fw_version,
        sign_image=getattr(args, 'sign_image', False),
        component=getattr(args, 'component_name', None),
        short_url=hasattr(args, 'use_short_url') and args.use_short_url,
        encrypt_payload=encrypt_payload
    )



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import abc
import base64
import hashlib
import logging
import re
import uuid
from collections import OrderedDict
from pathlib import Path
from typing import Optional
from typing import TypeVar

import yaml
from cryptography.hazmat.primitives.asymmetric import ec

from manifesttool.mtool.payload_format import PayloadFormat

READ_BLOCK_SIZE = 65536

ManifestAsnCodecBaseType = TypeVar(
    'ManifestAsnCodecBaseType', bound='ManifestAsnCodecBase')

logger = logging.getLogger('manifest-codec')

class ManifestAsnCodecBase(abc.ABC):

    @abc.abstractmethod
    def get_der_signed_resource(self, signature: bytes) -> bytes:
        raise NotImplementedError

    @abc.abstractmethod
    def get_signed_data(self) -> bytes:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def decode(
            cls,
            data: bytes,
            verification_key: Optional[ec.EllipticCurvePublicKey]
    ) -> OrderedDict:
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def get_name() -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def set_payload_version(self, version):
        raise NotImplementedError

    @abc.abstractmethod
    def set_update_priority(self, priority: int):
        raise NotImplementedError

    @abc.abstractmethod
    def set_vendor_id(self, vendor_id: bytes):
        raise NotImplementedError

    @abc.abstractmethod
    def set_vendor_data(self, data: bytes):
        raise NotImplementedError

    @abc.abstractmethod
    def set_class_id(self, class_id: bytes):
        raise NotImplementedError

    @abc.abstractmethod
    def set_device_id(self, device_id: bytes):
        raise NotImplementedError

    @abc.abstractmethod
    def set_payload_fingerprint(self, digest: bytes, size: int):
        raise NotImplementedError

    @abc.abstractmethod
    def set_payload_uri(self, uri: str):
        raise NotImplementedError

    @abc.abstractmethod
    def set_payload_format(self, payload_format: str):
        raise NotImplementedError

    @abc.abstractmethod
    def set_payload_metadata(
            self,
            installed_digest: bytes,
            installed_size: int,
            precursor_digest: bytes
    ):
        raise NotImplementedError

    @abc.abstractmethod
    def set_update_certificate(self, cert_data: bytes):
        raise NotImplementedError

    @abc.abstractmethod
    def set_image_signature(self, signature: bytes):
        raise NotImplementedError

    @abc.abstractmethod
    def set_component_name(self, component: str):
        raise NotImplementedError

    def process_input_config(self, fw_version, input_cfg: dict) -> bytes:
        self.set_payload_version(fw_version)

        if 'priority' in input_cfg:
            self.set_update_priority(int(input_cfg['priority']))
        elif self.get_name() == 'v3':
            raise AssertionError('priority filed must be provided')

        self.set_component_name(input_cfg.get('component', 'MAIN'))

        vendor_id = self._encode_vendor_cfg(input_cfg)

        self._encode_device_cfg(input_cfg, vendor_id)

        installed_digest = self._encode_payload_cfg(input_cfg)

        return installed_digest

    def _encode_payload_cfg(self, input_cfg) -> bytes:

        if 'payload' not in input_cfg:
            raise AssertionError('payload element not found')

        self._encode_payload_url(input_cfg)

        payload_format = self._encode_payload_format(input_cfg)

        # calc 'file-path' (local file) digest and size for patch,
        #  the local file is the generated patch
        # for encrypted-raw, the local file is the plain image
        local_digest, local_size = \
            self._calc_file_digest_and_size(input_cfg)

        if payload_format == PayloadFormat.RAW:
            payload_digest, payload_size = \
                local_digest, local_size
            installed_digest, installed_size, precursor_digest = \
                local_digest, local_size, None

        elif payload_format == PayloadFormat.PATCH:
            payload_digest, payload_size = \
                local_digest, local_size
            installed_digest, installed_size, precursor_digest = \
                self._get_delta_metadata(input_cfg)

        elif payload_format == PayloadFormat.ENCRYPTED_RAW:
            payload_digest, payload_size = \
                self._get_encryption_metadata(input_cfg)
            installed_digest, installed_size, precursor_digest = \
                local_digest, local_size, None

        else:
            raise AssertionError('Unsupported payload format')

        # set the downloaded payload fingerprint
        self.set_payload_fingerprint(
            digest=payload_digest,
            size=int(payload_size)
        )

        if payload_format == PayloadFormat.RAW:
            return payload_digest

        # set the installed payload fingerprint
        self.set_payload_metadata(
            installed_digest=installed_digest,
            installed_size=int(installed_size),
            precursor_digest=precursor_digest)

        return installed_digest

    @staticmethod
    def _calc_file_digest_and_size(input_cfg):
        if 'file-path' in input_cfg['payload']:
            file_path = input_cfg['payload']['file-path']
            # calc digest and size of file-path
            with open(file_path, 'rb') as fh:
                hash_ctx = hashlib.sha256()
                buf = fh.read(READ_BLOCK_SIZE)
                while buf:
                    hash_ctx.update(buf)
                    buf = fh.read(READ_BLOCK_SIZE)
                file_digest = hash_ctx.digest()
                file_size = fh.tell()
        else:
            raise AssertionError(
                'payload:file-path must be provided')
        return file_digest, file_size

    def _encode_payload_url(self, input_cfg):
        if 'url' in input_cfg['payload']:
            self.set_payload_uri(input_cfg['payload']['url'])
        else:
            raise AssertionError('payload:url not found')

    def _encode_payload_format(self, input_cfg):
        if 'format' in input_cfg['payload']:
            try:
                payload_format = PayloadFormat(input_cfg['payload']['format'])
            except KeyError as ex:
                raise AssertionError('unknown payload-format') from ex
            self.set_payload_format(str(payload_format.value))
        else:
            raise AssertionError('payload-format element not found')
        return payload_format

    @staticmethod
    def _get_delta_metadata(input_cfg):
        delta_file = Path(input_cfg['payload']['file-path'])
        delta_config_file = delta_file.with_suffix('.yaml')
        with delta_config_file.open('rb') as fh:
            delta_cfg = yaml.safe_load(fh)
        installed_digest = base64.b64decode(delta_cfg['installed-digest'])
        installed_size = int(delta_cfg['installed-size'])
        precursor_digest = \
            base64.b64decode(delta_cfg['precursor-digest'])
        return installed_digest, installed_size, precursor_digest

    @staticmethod
    def _get_encryption_metadata(input_cfg):
        if 'encrypted' not in input_cfg['payload']:
            raise AssertionError('payload:encrypted must be provided')
        if 'digest' not in input_cfg['payload']['encrypted']:
            raise AssertionError(
                'payload:encrypted digest must be provided')
        if 'size' not in input_cfg['payload']['encrypted']:
            raise AssertionError(
                'payload:encrypted size must be provided')

        encrypted_digest = bytearray.fromhex(
            input_cfg['payload']['encrypted']['digest']
        )
        encrypted_size = int(input_cfg['payload']['encrypted']['size'])

        return encrypted_digest, encrypted_size

    def _encode_device_cfg(self, input_cfg, vendor_id):
        if 'device' not in input_cfg:
            raise AssertionError('device element not found')
        if 'model-name' in input_cfg['device']:
            class_id = uuid.uuid5(vendor_id, input_cfg['device']['model-name'])
        elif 'class-id' in input_cfg['device']:
            class_id = uuid.UUID(input_cfg['device']['class-id'])
        else:
            raise AssertionError(
                'either device:class-id or device:model-name must be provided')
        logger.info('Class-ID: %s', class_id.bytes.hex())
        self.set_class_id(class_id.bytes)

    def _encode_vendor_cfg(self, input_cfg):
        if 'vendor' not in input_cfg:
            raise AssertionError('invalid vendor element not found')
        if 'domain' in input_cfg['vendor']:
            if not re.match(r'^\S+\.\S+$', input_cfg['vendor']['domain']):
                raise AssertionError('invalid vendor_domain')
            vendor_domain = input_cfg['vendor']['domain']
            vendor_id = uuid.uuid5(uuid.NAMESPACE_DNS, vendor_domain)
        elif 'vendor-id' in input_cfg['vendor']:
            vendor_id = uuid.UUID(input_cfg['vendor']['vendor-id'])
        else:
            raise AssertionError(
                'either vendor:vendor-id or vendor:domain must be provided')

        logger.info('Vendor-ID: %s', vendor_id.bytes.hex())
        self.set_vendor_id(vendor_id.bytes)

        if 'vendor' in input_cfg and 'custom-data-path' in input_cfg['vendor']:
            self.set_vendor_data(
                Path(input_cfg['vendor']['custom-data-path']).read_bytes()
            )
        else:
            pass  # vendor data is an optional field
        return vendor_id



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

from setuptools import setup, find_packages, Extension

import manifesttool

armbsdiff = Extension(
    'manifesttool.armbsdiff',
    sources=[
        'bsdiff/bsdiff.c',
        'bsdiff/bsdiff_helper.c',
        'bsdiff/bsdiff_python.c',
        'bsdiff/lz4.c',
        'bsdiff/varint.c'
    ],
    include_dirs=['bsdiff'],
    define_macros=[('LZ4_MEMORY_USAGE', '10')],
    extra_compile_args=['--std=c99', '-O3']
)

with open('requirements.txt', 'rt') as fh:
    tool_requirements = fh.readlines()

setup(
    name='manifest-tool',
    version=manifesttool.__version__,
    description='Tool/lib to create and parse manifests',
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url='https://github.com/PelionIoT/manifest-tool',
    author='Pelion',
    author_email='pdm-support@pelion.com',
    license='Apache 2.0',
    packages=find_packages(exclude=['tests']),
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "manifest-tool=manifesttool.mtool.mtool:entry_point",
            "manifest-dev-tool=manifesttool.dev_tool.dev_tool:entry_point",
            "manifest-delta-tool=manifesttool.delta_tool.delta_tool:entry_point"
        ],
    },
    python_requires='>=3.6.0',
    include_package_data=True,
    install_requires=tool_requirements,
    ext_modules=[armbsdiff]
)



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import base64
import hashlib
import logging
import time
import urllib.parse
from pathlib import Path
from typing import NewType, List

import requests

FW_UPLOAD = '/v3/firmware-images'
FW_UPLOAD_JOBS = '/v3/firmware-images/upload-jobs'
FW_UPLOAD_JOB = '/v3/firmware-images/upload-jobs/{id}'
FW_UPLOAD_JOB_CHUNK = '/v3/firmware-images/upload-jobs/{id}/chunks'
FW_IMAGE = '/v3/firmware-images/{id}'

FW_MANIFESTS = '/v3/firmware-manifests'
FW_MANIFEST = '/v3/firmware-manifests/{id}'

FW_CAMPAIGNS = '/v3/update-campaigns'
FW_CAMPAIGN = '/v3/update-campaigns/{id}'
FW_CAMPAIGN_STOP = '/v3/update-campaigns/{id}/stop'
FW_CAMPAIGN_START = '/v3/update-campaigns/{id}/start'
FW_CAMPAIGN_STATISTICS = '/v3/update-campaigns/{id}/statistics'
FW_CAMPAIGN_STATISTICS_EVENTS = \
    '/v3/update-campaigns/{id}/statistics/{summary_id}/event_types'
FW_CAMPAIGN_DEV_METADATA = '/v3/update-campaigns/{id}/campaign-device-metadata'

CAMPAIGN_ACTIVE_PHASES = [
    'starting',
    'active'
]

CAMPAIGN_NOT_STARTED_PHASES = [
    'draft',
    'stopped'
]

FW_UPLOAD_MAX_SMALL_SIZE = int(100 * 1024 * 1024)

# must be greater than 5MB - AWS limitation
FW_UPLOAD_CHUNK_SIZE = int(6 * 1024 * 1024)

LOG = logging.getLogger('pelion')

URL = NewType('URL', str)
ID = NewType('ID', str)
Phase = NewType('Phase', str)

class UpdateServiceApi:
    def __init__(self, host: URL, access_key: str):
        """
        Create REST API provider for Update Service APIs
        :param host: Pelion service URL
        :param access_key: account access key
        """
        self._host = host
        self._default_headers = {
            'Authorization': 'Bearer {}'.format(access_key)
        }

    def _url(self, api, **kwargs) -> str:
        """
        Helper function for constructing full REST API URL

        Concatenates Pelion host URL with the desired REST API url and expands
        any patterns present in the URL (e.g. v3/manifests/{id})
        :param api: REST API URL part
        :param kwargs: dictionary for expanding the the URL patterns
        :return: full URL
        """
        if kwargs:
            return urllib.parse.urljoin(self._host, api.format(**kwargs))
        return urllib.parse.urljoin(self._host, api)

    def _headers(self, extra_headers: dict = None) -> dict:
        """
        Helper function for creating request header

        Initializes a dictionary with common headers and allows extending it
        with REST API specific data
        :param extra_headers: dictionary containing extra headers to be sent as
        part of REST API request
        :return: headers dictionary
        """
        copy = self._default_headers.copy()
        if extra_headers:
            copy.update(extra_headers)
        return copy

    @staticmethod
    def _chunk_md5(chunk: bytes) -> str:
        """
        Helper function for calculating MD5 digest over bytes
        :param chunk: bytes to be processed
        :return: base64 encoded MD5 digest
        """
        return base64.b64encode(hashlib.md5(chunk).digest()).decode()  # nosec

    @staticmethod
    def _print_upload_progress(current: int, total: int):
        """
        Helper function for printing upload progress
        :param current: bytes uploaded so far
        :param total: uploaded file size
        """
        if total < FW_UPLOAD_CHUNK_SIZE:
            return
        increments = 50
        progress = ((current / total) * 100)
        count = int(progress // (100 / increments))
        text = '\r[{0: <{1}}] {2:.2f}%'.format(
            '=' * count, increments, progress)
        print(text, end='\n' if progress == 100 else '')

    def fw_upload(self, fw_name: str, image: Path, encrypt: bool) \
            -> dict:
        """
        Upload FW image
        :param fw_name: update candidate image name as will appear on Pelion
               portal
        :param image: candidate FW image
        :param encrypt: request to encrypt FW image
        :return: uploaded image meta
        """
        fw_size = image.stat().st_size

        if fw_size < FW_UPLOAD_MAX_SMALL_SIZE:
            fw_meta = self._upload_small_image(
                fw_name, image, encrypt
            )
        else:
            fw_meta = self._upload_big_image(
                fw_name, fw_size, image, encrypt
            )

        LOG.info('Uploaded FW image %s', fw_meta['datafile'])
        return fw_meta

    def _upload_image_chunk(self, chunk: bytes, job_id: ID):
        """
        Helper function for uploading FW image chunk
        :param chunk: FW image chunk to be uploaded
        :param job_id: upload job ID
        """
        response = requests.post(
            self._url(FW_UPLOAD_JOB_CHUNK, id=job_id),
            headers=self._headers(
                {
                    'Content-Type': 'binary/octet-stream',
                    'Content-MD5': self._chunk_md5(chunk)
                }
            ),
            data=chunk
        )
        response.raise_for_status()

    def _upload_big_image(
        self,
        fw_name: str,
        fw_size: int,
        image: Path,
        encrypt: bool
    ) -> dict:
        """
        Helper function for uploading small FW image
        :param fw_name: fw name as appears on the portal
        :param image: candidate FW image
        :param encrypt: Request to encrypt FW image
        :return: uploaded image meta
        """
        job_id = None
        try:
            with image.open('rb') as fh:
                job_id = self._create_upload_job(fw_name, encrypt)
                self._print_upload_progress(0, fw_size)
                upload_counter = 0
                while True:
                    chunk = fh.read(FW_UPLOAD_CHUNK_SIZE)
                    upload_counter += len(chunk)
                    self._upload_image_chunk(chunk, job_id)
                    self._print_upload_progress(upload_counter, fw_size)
                    if not chunk:
                        break
            LOG.debug('FW upload job %s completed', job_id)
            return self._get_fw_image_meta(job_id)
        except requests.HTTPError:
            LOG.error('FW image upload failed')
            raise
        finally:
            if job_id:
                self._delete_upload_job(job_id)

    def _upload_small_image(
        self,
        fw_name: str,
        image: Path,
        encrypt: bool
    ) -> dict:
        """
        Helper function for uploading small FW image
        :param fw_name: fw name as appears on the portal
        :param image: candidate FW image
        :param encrypt: Request to encrypt FW image
        :return: uploaded image meta
        """
        with image.open('rb') as fh:
            response = requests.post(
                self._url(FW_UPLOAD),
                headers=self._headers(),
                files={
                    'datafile': (image.name, fh),
                },
                data={
                    'name': fw_name,
                    'datafile_encryption': encrypt
                }
            )
            response.raise_for_status()
            return response.json()

    def _get_fw_image_meta(self, job_id: ID) -> dict:
        """
        Helper function for extracting uploaded image URL and ID
        :param job_id: upload job ID
        :return: uploaded image meta
        """
        response = requests.get(
            self._url(FW_UPLOAD_JOB, id=job_id),
            headers=self._headers())
        response.raise_for_status()
        image_id = response.json()['firmware_image_id']
        response = requests.get(
            self._url(FW_IMAGE, id=image_id),
            headers=self._headers())
        response.raise_for_status()

        return response.json()

    def _delete_upload_job(self, job_id: ID):
        """
        Helper function for deleting a firmware image upload job
        :param job_id: upload job ID
        """
        response = requests.delete(
            self._url(FW_UPLOAD_JOB, id=job_id),
            headers=self._headers()
        )
        response.raise_for_status()
        LOG.debug('FW upload job %s deleted', job_id)

    def _create_upload_job(self, fw_name: str, encrypt: bool) -> ID:
        """
        Helper function for creating a firmware image upload job
        :param fw_name: fw name as appears on the portal
        :param encrypt: Request to encrypt FW image
        :return: upload job ID
        """
        response = requests.post(
            self._url(FW_UPLOAD_JOBS),
            headers=self._headers(),
            json={
                'name': fw_name,
                'datafile_encryption': encrypt
            }
        )
        response.raise_for_status()
        response_data = response.json()
        job_id = response_data['id']
        LOG.debug('FW upload job created %s', job_id)
        return job_id

    def fw_delete(self, image_id: ID):
        """
        Delete candidate image from Pelion portal
        :param image_id: image ID as appears on the portal
        """
        response = requests.delete(
            self._url(FW_IMAGE, id=image_id),
            headers=self._headers()
        )
        response.raise_for_status()
        LOG.info('Deleted FW image %s', image_id)

    def manifest_upload(self, name: str, manifest: Path) -> ID:
        """
        Upload manifest file to Pelion service
        :param name: manifest name as will appear on Pelion portal
        :param manifest: manifest file
        :return: manifest ID as reported by portal
        """
        try:
            with manifest.open('rb') as fh:
                response = requests.post(
                    self._url(FW_MANIFESTS),
                    headers=self._headers(),
                    files={
                        'datafile': (manifest.name, fh),
                    },
                    data={
                        'name': name
                    }
                )
                response.raise_for_status()
                manifest_id = response.json()['id']
                LOG.info('Uploaded manifest ID: %s', manifest_id)
                return manifest_id
        except requests.HTTPError:
            LOG.error('Failed to upload manifest')
            raise

    def manifest_delete(self, manifest_id: ID):
        """
        Delete manifest file from Pelion portal
        :param manifest_id: manifest ID to be deleted
        """
        response = requests.delete(
            self._url(FW_MANIFEST, id=manifest_id),
            headers=self._headers()
        )
        response.raise_for_status()
        LOG.info('Deleted manifest ID: %s', manifest_id)

    def campaign_create(
            self,
            name: str,
            manifest_id: ID,
            device_filter: str
    ) -> ID:
        """
        Create update campaign on Pelion portal

        :param name: campaign name as will appear on Pelion portal
        :param manifest_id: manifest ID as
        :param device_filter: device filter query
        :return: campaign ID as reported by portal
        """
        try:
            response = requests.post(
                self._url(FW_CAMPAIGNS),
                headers=self._headers(),
                json={
                    'campaign_strategy': 'one-shot',
                    'description': 'Development campaign',
                    'device_filter': device_filter,
                    'name': name,
                    'root_manifest_id': manifest_id
                }
            )
            response.raise_for_status()
            campaign_id = response.json()['id']
            LOG.info('Created Campaign ID: %s', campaign_id)
            return campaign_id
        except requests.HTTPError:
            LOG.error('Failed to create campaign')
            raise

    def campaign_delete(self, campaign_id: ID):
        """
        Delete inactive (stopped/draft) update campaign
        :param campaign_id: campaign ID
        """
        response = requests.delete(
            self._url(FW_CAMPAIGN, id=campaign_id),
            headers=self._headers()
        )
        response.raise_for_status()
        LOG.info('Deleted campaign %s', campaign_id)

    def campaign_stop(self, campaign_id: ID):
        """
        Stop update campaign
        :param campaign_id: campaign ID
        """
        try:
            # check campaign phase and skip if it not active
            curr_phase = self.campaign_get(campaign_id)['phase']
            if not self.campaign_is_active(curr_phase):
                return
            # send request to stop
            response = requests.post(
                self._url(FW_CAMPAIGN_STOP, id=campaign_id),
                headers=self._headers()
            )
            response.raise_for_status()
            curr_phase = self.campaign_get(campaign_id)['phase']
            # The campaign is stopping. wait a minute for it to finish
            retries = 60
            while curr_phase == 'stopping' and retries > 0:
                time.sleep(1)
                curr_phase = self.campaign_get(campaign_id)['phase']
                retries -= 1
            if retries == 0:
                LOG.debug('Stopping Campaign timed out')
            LOG.info('Stopped campaign %s', campaign_id)
        except requests.HTTPError as ex:
            # ignore error from stop campaign as it may already be stopped
            LOG.debug("Failed stopping campaign - %s", ex)

    def campaign_start(self, campaign_id: ID):
        """
        Start update campaign
        :param campaign_id: campaign ID
        """
        response = requests.post(
            self._url(FW_CAMPAIGN_START, id=campaign_id),
            headers=self._headers()
        )
        response.raise_for_status()

    def campaign_get(self, campaign_id: ID) -> dict:
        """
        Get update campaign current representation/state
        :param campaign_id:
        :return: Dictionary with current update campaign state
        """
        response = requests.get(
            self._url(FW_CAMPAIGN, id=campaign_id),
            headers=self._headers()
        )
        response.raise_for_status()
        return response.json()

    def campaign_statistics(self, campaign_id: ID) -> List[dict]:
        """
        Get campaign statistics

        :param campaign_id: campaign ID
        :return: List of statistics for a campaign
        """
        try:
            response = requests.get(
                self._url(FW_CAMPAIGN_STATISTICS, id=campaign_id),
                headers=self._headers()
            )
            response.raise_for_status()
            return response.json()['data']
        except requests.HTTPError:
            pass
        return []

    def campaign_statistic_events(self,
                                  campaign_id: ID,
                                  summary_id: str) -> List[dict]:
        """
        Get campaign events grouped by summary

        :param campaign_id: campaign ID
        :param summary_id: The summary status
            Available values: fail, success, info, skipped
        :return: List of statistics for a campaign
        """
        try:
            response = requests.get(
                self._url(FW_CAMPAIGN_STATISTICS_EVENTS,
                          id=campaign_id, summary_id=summary_id),
                headers=self._headers()
            )
            response.raise_for_status()
            return response.json()['data']
        except requests.HTTPError:
            pass
        return []

    def campaign_device_metadata(self, campaign_id: ID) -> List[dict]:
        """
        Get metadata for devices participating in update campaign

        Note: this function and assumes small number of devcises participating
        in update campaign (as part of developer flow). Thus it does not
        handles pagination
        :param campaign_id: campaign ID
        :return: List of device metadata dictionaries for devices participating
        in update campaign
        """
        response = requests.get(
            self._url(FW_CAMPAIGN_DEV_METADATA, id=campaign_id),
            headers=self._headers()
        )
        response.raise_for_status()
        return response.json()['data']

    @staticmethod
    def campaign_is_active(phase: Phase) -> bool:
        """
        Helper function for understanding if update campaign is in active phase
        :param phase: current update campaign phase
        :return: True if update campaign phase is starting or active
        """
        return phase in CAMPAIGN_ACTIVE_PHASES

    @staticmethod
    def campaign_is_not_started(phase: Phase) -> bool:
        """
        Helper function for understanding if update campaign is not
        started phase
        :param phase: current update campaign phase
        :return: True if update campaign phase is starting or active
        """
        return phase in CAMPAIGN_NOT_STARTED_PHASES
